#include "config.h"
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <glob.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>
#include <limits.h>

#include "saphire.h"
#include "saphire_inner.h"

// スペースを飛ばす
static void skip_spaces(char** p, string_obj* not_evaled)
{
    while(**p == ' ' || **p == '\t') {
        string_push_back2(not_evaled, **p);
        (*p)++;
    }
}

// {}, <>, (), []内を丸ごと取りたい関数。ネストしていてもOK
// 中身はもう一度parseされる
// 変数展開も対応
static BOOL read_until_next_close_character(char first_char, char last_char, char** p, string_obj* not_evaled, int* read_end_of_statment, BOOL line_inc, char* sname, int* sline)
{
    if(**p == first_char) {
        string_push_back2(not_evaled, **p);
        (*p)++;

        int nest = 0;
        BOOL squote = FALSE;
        BOOL dquote = FALSE;
        while(1) {
            /// クォート ///
            if(!squote && !dquote && **p == '\\') {
                string_push_back2(not_evaled, **p);
                (*p)++;

                if(line_inc && **p == '\n') {
                    (*sline)++;
                }
                else if(**p == 0) {
                    err_msg("unexpected end(0) after quote", sname, *sline);
                    return FALSE;
                }

                string_push_back2(not_evaled, **p);
                (*p)++;
            }
            /// シングルクォート ///
            else if(!dquote && **p == '\'') {
                squote = !squote;
                string_push_back2(not_evaled, **p);
                (*p)++;
            }
            /// ダブルクォート ///
            else if(!squote && **p == '"') {
                dquote = !dquote;
                string_push_back2(not_evaled, **p);
                (*p)++;
            }
            /// コマンドライン展開
            else if(!squote && **p == '$') {
                // コマンド展開 $()
                if(*(*p+1) == '$' && *(*p+2) == '(' || *(*p+1) == '(')
                {
                    if(*(*p+1) == '$') {
                        string_push_back2(not_evaled, **p); // $
                        (*p)++;
                        string_push_back2(not_evaled, **p); // $
                        (*p)++;
                    }
                    else {
                        string_push_back2(not_evaled, **p); // $
                        (*p)++;
                    }

                    if(!read_until_next_close_character('(', ')', p, not_evaled, read_end_of_statment, line_inc, sname, sline)) {
                        return FALSE;
                    }

                    *read_end_of_statment = 1;
                }
                else {
                    string_push_back2(not_evaled, **p);
                    (*p)++;
                }
            }
            else if(squote || dquote) {
                if(**p == 0) {
                    err_msg("require ' or \" before the end of source", sname, *sline);
                    return FALSE;
                }
                else {
                    if(line_inc && **p == '\n') {
                        (*sline)++;
                    }

                    string_push_back2(not_evaled, **p);
                    (*p)++;
                }
            }
            else if(**p == first_char) {
                nest++;
                string_push_back2(not_evaled, **p);
                (*p)++;
            }
            else if(**p == last_char) {
                string_push_back2(not_evaled, **p);
                (*p)++;

                if(nest == 0) {
                    break;
                }
                nest--;
            }
            else if(**p == 0) {
                char buf[256];
                snprintf(buf, 256, "read_util_next_close_character: unexpected end(0). need last char(%c)", last_char);
                err_msg(buf, sname, *sline);
                return FALSE;
            }
            else if(line_inc && **p == '\n') {
                (*sline)++;
                string_push_back2(not_evaled, **p);
                (*p)++;
            }
            else {
                string_push_back2(not_evaled, **p);
                (*p)++;
            }
        }
    }
    
    return TRUE;
}

// ブロックを取りたい関数 contentsに入る。
// 中身はもう一回パースされる
// コマンド展開は対応しない
static BOOL read_block(char** p, string_obj* contents, vector_obj* args, string_obj* not_evaled, BOOL line_inc, char* sname, int* sline)
{
    int line_of_head = *sline;
    string_push_back2(not_evaled, **p); //  {
    (*p)++;
    
    while(1) {
        if(**p == ' ') {
            string_push_back2(not_evaled, **p);
            string_push_back2(contents, *((*p)++));
        }
        else if(**p == '\t') {
            string_push_back2(not_evaled, **p);
            string_push_back2(contents, *((*p)++));
        }
        else if(**p == '\n') {
            line_inc && (*sline)++;
            string_push_back2(not_evaled, **p);
            string_push_back2(contents, *((*p)++));
        }
        else {
            break;
        }
    }

    /// 引数
    if(args) {
        if(**p == ':') {
            string_push_back2(not_evaled, **p);
            (*p)++;

            while(**p) {
                string_obj* str = STRING_NEW("");
                skip_spaces(p, not_evaled);
                while(**p >= 'a' && **p <= 'z'
                    || **p >= 'A' && **p <= 'Z'
                    || **p == '_')
                {
                    string_push_back2(str, **p);
                    string_push_back2(not_evaled, **p);
                    (*p)++;
                }
                skip_spaces(p, not_evaled);

                vector_add(args, str);

                if(**p >= 'a' && **p <= 'z'
                    || **p >= 'A' && **p <= 'Z'
                    || **p == '_')
                {
                }
                else {
                    if(**p == ':') {
                        string_push_back2(not_evaled, **p);
                        (*p)++;
                        break;
                    }
                    else {
                        char buf[1024];
                        snprintf(buf, 1024, "unexpected char(%c)", **p);
                        err_msg(buf, sname, line_of_head);
                        return FALSE;
                    }
                }
            }
        }

        while(1) {
            if(**p == ' ') {
                string_push_back2(not_evaled, **p);
                string_push_back2(contents, *((*p)++));
            }
            else if(**p == '\t') {
                string_push_back2(not_evaled, **p);
                string_push_back2(contents, *((*p)++));
            }
            else if(**p == '\n') {
                line_inc && (*sline)++;
                string_push_back2(not_evaled, **p);
                string_push_back2(contents, *((*p)++));
            }
            else {
                break;
            }
        }
    }

    BOOL squote = FALSE;
    BOOL dquote = FALSE;
    int nest = 0;
    while(1) {
        /// クォート
        if(!squote && !dquote && **p == '\\') {
            string_push_back2(not_evaled, **p);
            string_push_back2(contents, *((*p)++));

            if(line_inc && **p == '\n') {
                (*sline)++;
            }
            else if(*p == 0) {
                err_msg("unsigned end after quote",sname, *sline);
                return FALSE;
            }

            string_push_back2(not_evaled, **p);
            string_push_back2(contents, *((*p)++));
        }
        /// シングルクォート
        else if(!dquote && **p == '\'') {
            string_push_back2(not_evaled, **p);
            string_push_back2(contents, *((*p)++));
            squote = !squote;
        }
        /// ダブルクォート
        else if(!squote && **p == '"') {
            string_push_back2(not_evaled, **p);
            string_push_back2(contents, *((*p)++));
            dquote = !dquote;
        }
        /// シングルクォート、ダブルクォート中 ///
        else if(squote || dquote) {
            if(**p == 0) {
                err_msg("get block: require \" or \'", sname, line_of_head);
                return FALSE;
            }
            else if(line_inc && **p == '\n') {
                (*sline)++;
            }

            string_push_back2(not_evaled, **p);
            string_push_back2(contents, *((*p)++));
        }
        else if(**p == '{') {
            string_push_back2(not_evaled, **p);
            string_push_back2(contents, *((*p)++));
            nest++;
        }
        else if(**p == '}') {
            string_push_back2(not_evaled, **p);
            (*p)++;
            nest--;

            if(nest == -1) {
                break;
            }
            else {
                string_push_back2(contents, '}');
            }
        }
        else if(**p == 0) {
            err_msg("get block: found 0. require }", sname, line_of_head);
            return FALSE;
        }
        else if(**p == '\n') {
            line_inc && (*sline)++;
            string_push_back2(not_evaled, **p);
            string_push_back2(contents, **p);
            (*p)++;
        }
        else {
            string_push_back2(not_evaled, **p);
            string_push_back2(contents, **p);
            (*p)++;
        }
    }

    return TRUE;
}

/// チルダ展開
static void expand_tilda(string_obj* buf, string_obj* path)
{
    char* p = string_c_str(buf);
    p++; // ~

    string_obj* user = STRING_NEW("");
    while(*p && *p != '/') {
        string_push_back2(user, *p);
        p++;
    }

    /// ユーザー名からホームディレクトリに展開
    if(string_length(user) == 0) {
        char* home = getenv("HOME");
        if(home) {
            string_push_back(path, home);
        }
        else {
            struct passwd* pw = getpwuid(getuid());
            if(pw) {
                string_push_back(path, pw->pw_dir);
            }
            else {
                string_push_back(path, "~");
            }
        }
    }
    else {
        struct passwd* pw = getpwnam(string_c_str(user));
        if(pw) {
            string_push_back(path, pw->pw_dir);
        }
        else {
            string_push_back2(path, '~');
            string_push_back(path, string_c_str(user));
        }
    }

    /// 残りを代入
    string_push_back(path, p);
}

/// 一単語読み込み
static BOOL read_one_argument(char** p, string_obj* buf, string_obj* not_evaled, int* read_end_of_statment, sCommand* command, char* sname, int* sline)
{
    /// 単語の区切り記号かどうかのテーブル 127以下なら
    static unsigned char table[] = { 
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
        1, 0, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 
    };

    BOOL squote = FALSE;
    BOOL dquote = FALSE;
    while(**p) {
        if(gKitutukiSigInt) {
            gKitutukiSigInt = FALSE;
            err_msg("signal interrupt", sname, *sline);

            return FALSE;
        }

        /// 空文字
        if(!squote && !dquote && **p == '"' && *(*p+1) == '"') {
            break;
        }
        else if(!dquote && !squote && **p =='\'' && *(*p+1) == '\'') {
            break;
        }
        else if(**p == '\\' && *(*p+1) == 't') {
            string_push_back2(not_evaled, **p);
            string_push_back2(not_evaled, *(*p+1));

            string_push_back2(buf, '\t');

            (*p)+=2;
        }
        else if(**p == '\\' && *(*p+1) == 'n') {
            string_push_back2(not_evaled, **p);
            string_push_back2(not_evaled, *(*p+1));

            string_push_back2(buf, '\n');

            (*p)+=2;
        }
        else if(**p == '\\' && *(*p+1) == 'r') {
            string_push_back2(not_evaled, **p);
            string_push_back2(not_evaled, *(*p+1));

            string_push_back2(buf, '\r');

            (*p)+=2;
        }
        else if(**p == '\\' && *(*p+1) == 'a') {
            string_push_back2(not_evaled, **p);
            string_push_back2(not_evaled, *(*p+1));

            string_push_back2(buf, '\a');

            (*p)+=2;
        }
        else if(!squote && !dquote && **p == '\\') {
            string_push_back2(not_evaled, **p);
            (*p)++;

            if(**p == 0) {
                err_msg("unexpected end. can't quote null.", sname, *sline);
                return FALSE;
            }
            else if(**p == '\n') {
                (*sline)++;
            }

            string_push_back2(not_evaled, **p);
            string_push_back2(buf, **p);

            (*p)++;
        }
        else if(!dquote && **p == '\'') {
            string_push_back2(not_evaled, **p);
            (*p)++;
            squote = !squote;
        }
        else if(!squote && **p == '"') {
            dquote = !dquote;
            string_push_back2(not_evaled, **p);
            (*p)++;
        }
        /// コマンドライン展開
        else if(!squote && **p == '$') {
            // コマンド展開 $()
            if(*(*p+1) == '(' || *(*p+1) == '$' && *(*p+2) == '(')
            {
                if(*(*p+1) == '$') {
                    string_push_back2(not_evaled, **p); // $
                    (*p)++;
                    string_push_back2(not_evaled, **p); // (
                    (*p)++;
                }
                else {
                    string_push_back2(not_evaled, **p); // $
                    (*p)++;
                }

                if(!read_until_next_close_character('(', ')', p, not_evaled, read_end_of_statment, TRUE, sname, sline)) {
                    return FALSE;
                }

                *read_end_of_statment = 1;
            }
            else {
                string_push_back2(not_evaled, **p);
                string_push_back2(buf, **p);

                (*p)++;
            }
        }
        /// シングルクォート／ダブルクォート中
        else if(squote || dquote) {
            if(**p == 0) {
                err_msg("close single quote or double quote", sname, *sline);
                return FALSE;
            }
            else if(**p == '\n') {
                (*sline)++;
            }

            string_push_back2(not_evaled, **p);
            string_push_back2(buf, **p);
            (*p)++;
        }
        /// チルダ展開 ///
        else if(**p == '~') {
            string_push_back2(not_evaled, **p);
            string_push_back2(buf, **p);
            (*p)++;

            if(strlen(string_c_str(buf)) == 1 && command) {
                if(!read_one_argument(p, buf, not_evaled, read_end_of_statment, NULL, sname, sline))
                {
                    return FALSE;
                }

                string_obj* path = STRING_NEW("");
                expand_tilda(buf, path);
                vector_add(command->mArgs, sArg_new(0, path));
                string_put(buf, ""); // チルダ展開のパターンはmArgsに追加しない
                break;
            }
        }
        /// グロブ ///
        else if(command && vector_size(command->mArgs) > 0 && (**p == '[' || **p == '?' || **p == '*')) {
            string_push_back2(not_evaled, **p);
            string_push_back2(buf, **p);
            (*p)++;

            if(command) {
                if(!read_one_argument(p, buf, not_evaled, read_end_of_statment, NULL, sname, sline))
                {
                    return FALSE;
                }

                glob_t result;
                int rc = glob(string_c_str(buf), 0, NULL, &result);
                if(rc == GLOB_NOSPACE) {
                    err_msg("read_one_argument: out of space during glob operation", sname, *sline);
                    return FALSE;
                }
                else if(rc == GLOB_NOMATCH) {
                }
                else {
                    int i;
                    for(i=0; i<result.gl_pathc; i++) {
                        char* file = result.gl_pathv[i];
                        if(strcmp(file, ".") != 0 && strcmp(file, "..") != 0)
                        {
                            vector_add(command->mArgs, sArg_new(0, STRING_NEW(file)));
                        }
                    }

                    string_put(buf, ""); // グロブのパターンはmArgsに追加しない
                }
                break;
            }
        }
        // ->は下で処理
        else if(command && vector_size(command->mArgs) == 0 && **p == '-' && *(*p+1) == '>') {
            string_push_back2(not_evaled, **p);
            string_push_back2(buf, **p);

            (*p)++;

            string_push_back2(not_evaled, **p);
            string_push_back2(buf, **p);

            (*p)++;
        }
        /// utf-8キャラクター
        else if(((unsigned char)**p) > 127) {
            string_push_back2(not_evaled, **p);
            string_push_back2(buf, **p);

            (*p)++;
        }
        // クィックコマンド展開は下で処理
        else if(**p == '@' && *(*p+1) == '(' || **p == '@' && *(*p+1) == '@' && *(*p+2) == '(') {
            break;
        }
        // % のりダイレクトは下で処理
        else if(**p == '%' && (*(*p+1) == '>' || *(*p+1) == '2')) {
            break;
        }
        else if(table[**p]) {
            break;
        }
        else {
            string_push_back2(not_evaled, **p);
            string_push_back2(buf, **p);

            (*p)++;
        }
    }

    return TRUE;
}

/// if文読み込み
static BOOL read_command_parse_if_statment(char** p, sCommand* command, sStatment* statment, string_obj* not_evaled, int* read_end_of_statment, char* sname, int* sline)
{
    skip_spaces(p, not_evaled);

    if(**p != '(') {
        err_msg("expected (", sname, *sline);
        return FALSE;
    }

    /// 条件式読み込み ///
    string_obj* str = STRING_NEW("");
    int read_end_of_statment2 = -1;
    if(!read_until_next_close_character('(', ')', p, str, &read_end_of_statment2, FALSE, sname, sline)) {
        return FALSE;
    }
    string_push_back(not_evaled, string_c_str(str));

    string_erase(str, 0, 1);
    string_trunc(str, string_length(str)-1);

    sStatments* condition = STATMENTS_NEW();

    if(!parse(string_c_str(str), sname, sline, condition))
    {

        return FALSE;
    }


    skip_spaces(p, not_evaled);

    /// 実行文 ///
    sStatments* loop = STATMENTS_NEW();

    if(**p == '{') {
        string_obj* str = STRING_NEW("");
        if(!read_block(p, str, NULL, not_evaled, FALSE, sname, sline)) {
            return FALSE;
        }

        if(!parse(string_c_str(str), sname, sline, loop))
        {
            return FALSE;
        }
    }
    else {
        err_msg("expected block.", sname, *sline);
        return FALSE;
    }

    skip_spaces(p, not_evaled);

    vector_obj* conditions = VECTOR_NEW(30);
    vector_obj* loops = VECTOR_NEW(30);
    
    vector_add(conditions, condition);
    vector_add(loops, loop);

    /// elif 開始 ///
    if(**p == 'e' && *(*p+1) == 'l' 
        && *(*p+2) == 'i' && *(*p+3) == 'f' 
        && (*(*p+4) == ' ' || *(*p+4) == '\t' || *(*p+4) == '('))
    {
        while(**p == 'e' && *(*p+1) == 'l' && *(*p+2) == 'i' 
            && *(*p+3) == 'f' 
            && (*(*p+4) == ' ' || *(*p+4) == '\t' || *(*p+4) == '('))
        {
            string_push_back2(not_evaled, **p);
            (*p)++;
            string_push_back2(not_evaled, **p);
            (*p)++;
            string_push_back2(not_evaled, **p);
            (*p)++;
            string_push_back2(not_evaled, **p);
            (*p)++;

            skip_spaces(p, not_evaled);

            if(**p != '(') {
                err_msg("expected (", sname, *sline);
                return FALSE;
            }

            /// 条件式 ///
            string_obj* str = STRING_NEW("");
            int read_end_of_statment = -1;
            if(!read_until_next_close_character('(', ')', p, str, &read_end_of_statment, FALSE, sname, sline)) {
                return FALSE;
            }
            string_push_back(not_evaled, string_c_str(str));

            sStatments* condition = STATMENTS_NEW();
            if(!parse(string_c_str(str), sname, sline, condition))
            {
                return FALSE;
            }

            vector_add(conditions, condition);

            skip_spaces(p, not_evaled);

            /// 実行文 ///
            sStatments* loop = STATMENTS_NEW();

            if(**p == '{') {
                string_obj* str = STRING_NEW("");
                if(!read_block(p, str, NULL, not_evaled, FALSE, sname, sline)) {
                    return FALSE;
                }

                if(!parse(string_c_str(str), sname, sline, loop))
                {
                    return FALSE;
                }
            }
            else {
                err_msg("expected block.", sname, *sline);
                return FALSE;
            }

            skip_spaces(p, not_evaled);

            vector_add(loops, loop);
        }
    }

    /// else 開始 ///
    if(**p == 'e' && *(*p+1) == 'l' && *(*p+2) == 's' && *(*p+3) == 'e' && (*(*p+4) == ' ' || *(*p+4) == '\t' || *(*p+4) == '('))
    {
        string_push_back2(not_evaled, **p);
        (*p)++;
        string_push_back2(not_evaled, **p);
        (*p)++;
        string_push_back2(not_evaled, **p);
        (*p)++;
        string_push_back2(not_evaled, **p);
        (*p)++;

        skip_spaces(p, not_evaled);
        
        /// 実行文 ///
        sStatments* loop = STATMENTS_NEW();

        if(**p == '{') {
            string_obj* str = STRING_NEW("");
            if(!read_block(p, str, NULL, not_evaled, FALSE, sname, sline)) {
                return FALSE;
            }

            if(!parse(string_c_str(str), sname, sline, loop))
            {
                return FALSE;
            }
        }
        else {
            err_msg("expected block.", sname, *sline);
            return FALSE;
        }

        skip_spaces(p, not_evaled);

        vector_add(loops, loop);
    }
    
    command->mKind = kIf;
    command->mExtra = sIf_new(conditions, loops);

    return TRUE;
}

/// try文読み込み
static BOOL read_command_parse_try_statment(char** p, sCommand* command, sStatment* statment, string_obj* not_evaled, int* read_end_of_statment, char* sname, int* sline)
{
    skip_spaces(p, not_evaled);

    /// 文読み込み ///
    sStatments* loops = STATMENTS_NEW();

    if(**p == '{') {
        string_obj* str = STRING_NEW("");
        if(!read_block(p, str, NULL, not_evaled, FALSE, sname, sline)) {
            return FALSE;
        }

        if(!parse(string_c_str(str), sname, sline, loops))
        {
            return FALSE;
        }
    }
    else {
        err_msg("expected block.", sname, *sline);
        return FALSE;
    }

    skip_spaces(p, not_evaled);

    sStatments* loops2 = STATMENTS_NEW();
    if(strstr(*p, "catch") == *p) {
        string_push_back2(not_evaled, **p);
        (*p)++;
        string_push_back2(not_evaled, **p);
        (*p)++;
        string_push_back2(not_evaled, **p);
        (*p)++;
        string_push_back2(not_evaled, **p);
        (*p)++;
        string_push_back2(not_evaled, **p);
        (*p)++;

        skip_spaces(p, not_evaled);

        if(**p == '{') {
            string_obj* str = STRING_NEW("");
            if(!read_block(p, str, NULL, not_evaled, FALSE, sname, sline)) {
                return FALSE;
            }

            if(!parse(string_c_str(str), sname, sline, loops2))
            {
                return FALSE;
            }
        }
        else {
            err_msg("expected block.", sname, *sline);
            return FALSE;
        }
    }
    else {
        err_msg("expected catch word", sname, *sline);

    }

    command->mKind = kTry;
        // これが無いとmNotEvaledの場合（変数がある場合）
        // 正しくcommandが消されない
    command->mExtra = sTry_new(loops, loops2);

    return TRUE;
}

/// while文読み込み
static BOOL read_command_parse_while_statment(char** p, sCommand* command, sStatment* statment, string_obj* not_evaled, int* read_end_of_statment, char* sname, int* sline)
{
    skip_spaces(p, not_evaled);

    if(**p != '(') {
        err_msg("expected (", sname, *sline);
        return FALSE;
    }

    /// 条件式読み込み ///
    string_obj* str = STRING_NEW("");
    int read_end_of_statment2 = -1;
    if(!read_until_next_close_character('(', ')', p, str, &read_end_of_statment2, FALSE, sname, sline)) {
        return FALSE;
    }
    string_push_back(not_evaled, string_c_str(str));

    string_erase(str, 0, 1);
    string_trunc(str, string_length(str)-1);

    sStatments* conditions = STATMENTS_NEW();
    if(!parse(string_c_str(str), sname, sline, conditions))
    {
        return FALSE;
    }

    while(**p) {
        if(**p == '\n') {
            (*sline)++;
            string_push_back2(not_evaled, **p);
            (*p)++;
        }
        else if(**p == ' ' || **p == '\t') {
            string_push_back2(not_evaled, **p);
            (*p)++;
        }
        else {
            break;
        }
    }

    /// ループ文読み込み ///
    sStatments* loops = STATMENTS_NEW();

    if(**p == '{') {
        string_obj* str = STRING_NEW("");
        if(!read_block(p, str, NULL, not_evaled, FALSE, sname, sline)) {
            return FALSE;
        }

        if(!parse(string_c_str(str), sname, sline, loops))
        {
            return FALSE;
        }
    }
    else {
        err_msg("expected block.", sname, *sline);
        return FALSE;
    }

    command->mKind = kWhile;
        // これが無いとmNotEvaledの場合（変数がある場合）
        // 正しくcommandが消されない
    command->mExtra = sWhile_new(conditions, loops);

    return TRUE;
}

/// def文読み込み
static BOOL read_command_parse_def_statment(char** p, sCommand* command, sStatment* statment, string_obj* not_evaled, int* read_end_of_statment, char* sname, int* sline)
{
    skip_spaces(p, not_evaled);

    /// 関数名を得る ///
    string_obj* name = STRING_NEW("");
    while(**p >= 'a' && **p <= 'z' 
            || **p >= 'A' && **p <= 'Z'
            || **p == '_' 
            || **p >= '0' && **p <= '9')
    {
        string_push_back2(not_evaled, **p);
        string_push_back2(name, **p);
        (*p)++;
    }

    /// 引数名指定 ///
    string_obj* arg_name = STRING_NEW("");
    if(**p == '(') {
        string_push_back2(not_evaled, **p);
        (*p)++;

        skip_spaces(p, not_evaled);

        while(**p >= 'a' && **p <= 'z' 
            || **p >= 'A' && **p <= 'Z'
            || **p == '_' 
            || **p >= '0' && **p <= '9')
        {
            string_push_back2(not_evaled, **p);
            string_push_back2(arg_name, **p);
            (*p)++;
        }

        skip_spaces(p, not_evaled);

        if(**p != ')') {
            err_msg("require )", sname, *sline);

            return FALSE;
        }

        string_push_back2(not_evaled, **p);
        (*p)++;
    }
/*
    /// 引数を取る ///
    vector_obj* args = VECTOR_NEW(10);
    string_obj* arg_name = STRING_NEW("");
    if(**p == '(') {
        string_push_back2(not_evaled, **p);
        (*p)++;
        while(1) {
            skip_spaces(p, not_evaled);

            string_obj* arg = STRING_NEW("");

            while(**p >= 'a' && **p <= 'z' 
                || **p >= 'A' && **p <= 'Z'
                || **p == '_' 
                || **p >= '0' && **p <= '9')
            {
                string_push_back2(not_evaled, **p);
                string_push_back2(arg, **p);
                (*p)++;
            }

            skip_spaces(p, not_evaled);

            if(string_length(arg) > 0) {
                vector_add(args, arg);
            }

            if(**p == ')') {
                break;
            }
            else if(**p >= 'a' && **p <= 'z' 
                || **p >= 'A' && **p <= 'Z'
                || **p == '_' 
                || **p >= '0' && **p <= '9')
            {
            }
            else {
                err_msg("invalid character in argument", sname, *sline);
                return FALSE;
            }
        }

        string_push_back2(not_evaled, **p);
        (*p)++;
    }
*/

    while(**p) {
        if(**p == '\n') {
            (*sline)++;
            string_push_back2(not_evaled, **p);
            (*p)++;
        }
        else if(**p == ' ' || **p == '\t') {
            string_push_back2(not_evaled, **p);
            (*p)++;
        }
        else {
            break;
        }
    }

    string_obj* contents = STRING_NEW("");
    if(**p == '{') {
        if(!read_block(p, contents, NULL, not_evaled, FALSE, sname, sline)) {
            return FALSE;
        }
    }
    else {
        err_msg("unexpected token at def statment", sname, *sline);
        return FALSE;
    }

    sStatments* statments = STATMENTS_NEW();

    if(!parse(string_c_str(contents), sname, sline, statments)) 
    {

        return FALSE;
    }


    command->mKind = kDef;
        // これが無いとmNotEvaledの場合（変数がある場合）
        // 正しくcommandが消されない
    command->mExtra = sFunction_new(string_c_str(name), statments, string_c_str(arg_name), NULL);


    return TRUE;
}

/// class文読み込み
static BOOL read_command_parse_class_statment(char** p, sCommand* command, sStatment* statment, string_obj* not_evaled, int* read_end_of_statment, char* sname, int* sline)
{
    skip_spaces(p, not_evaled);

    /// クラス名を得る ///
    string_obj* name = STRING_NEW("");
    while(**p >= 'a' && **p <= 'z' 
            || **p >= 'A' && **p <= 'Z'
            || **p == '_' 
            || **p >= '0' && **p <= '9')
    {
        string_push_back2(not_evaled, **p);
        string_push_back2(name, **p);
        (*p)++;
    }

    while(**p) {
        if(**p == '\n') {
            (*sline)++;
            string_push_back2(not_evaled, **p);
            (*p)++;
        }
        else if(**p == ' ' || **p == '\t') {
            string_push_back2(not_evaled, **p);
            (*p)++;
        }
        else {
            break;
        }
    }

    string_obj* contents = STRING_NEW("");
    if(**p == '{') {
        if(!read_block(p, contents, NULL, not_evaled, FALSE, sname, sline)) {
            return FALSE;
        }
    }
    else {
        err_msg("unexpected token at def statment", sname, *sline);
        return FALSE;
    }

    sStatments* statments = STATMENTS_NEW();

    if(!parse(string_c_str(contents), sname, sline, statments)) 
    {

        return FALSE;
    }

    command->mKind = kClass;
        // これが無いとmNotEvaledの場合（変数がある場合）
        // 正しくcommandが消されない
    command->mExtra = sClass_new(string_c_str(name), statments, NULL);

    return TRUE;
}

/// １コマンド読み込み
static BOOL read_command(char** p, sCommand* command, sStatment* statment, string_obj* not_evaled, int* read_end_of_statment, BOOL block_termination, char* sname, int* sline)
{
    while(**p) {
        string_obj* buf = STRING_NEW("");

        if(!read_one_argument(p, buf, not_evaled, read_end_of_statment, command, sname, sline)) {
            return FALSE;
        }

        /// コマンドに追加 ///
        if(strcmp(string_c_str(buf), "") != 0) {
            if(vector_size(command->mArgs) == 0) {
                if(strcmp(string_c_str(buf), "if") == 0) {
                    vector_add(command->mArgs, sArg_new(0, buf));
                    string_push_back(statment->mTitle, string_c_str(buf));
                    string_push_back(statment->mTitle, " ");

                    if(!read_command_parse_if_statment(p, command, statment, not_evaled, read_end_of_statment, sname, sline))
                    {
                        return FALSE;
                    }
                }
                else if(strcmp(string_c_str(buf), "while") == 0) {
                    vector_add(command->mArgs, sArg_new(0, buf));
                    string_push_back(statment->mTitle, string_c_str(buf));
                    string_push_back(statment->mTitle, " ");

                    if(!read_command_parse_while_statment(p, command, statment, not_evaled, read_end_of_statment, sname, sline))
                    {
                        return FALSE;
                    }
                }
                else if(strcmp(string_c_str(buf), "try") == 0) {
                    vector_add(command->mArgs, sArg_new(0, buf));
                    string_push_back(statment->mTitle, string_c_str(buf));
                    string_push_back(statment->mTitle, " ");

                    if(!read_command_parse_try_statment(p, command, statment, not_evaled, read_end_of_statment, sname, sline))
                    {
                        return FALSE;
                    }
                }
                else if(strcmp(string_c_str(buf), "def") == 0) {
                    vector_add(command->mArgs, sArg_new(0, buf));
                    string_push_back(statment->mTitle, string_c_str(buf));
                    string_push_back(statment->mTitle, " ");

                    if(!read_command_parse_def_statment(p, command, statment, not_evaled, read_end_of_statment, sname, sline))
                    {
                        return FALSE;
                    }
                }
                else if(strcmp(string_c_str(buf), "class") == 0) {
                    vector_add(command->mArgs, sArg_new(0, buf));
                    string_push_back(statment->mTitle, string_c_str(buf));
                    string_push_back(statment->mTitle, " ");

                    if(!read_command_parse_class_statment(p, command, statment, not_evaled, read_end_of_statment, sname, sline))
                    {
                        return FALSE;
                    }
                }
                else {
                    if(vector_size(statment->mCommands) == 0 && strcmp(string_c_str(buf), "!") == 0)
                    {
                        statment->mRCodeReverse = TRUE;
                    }

                    vector_add(command->mArgs, sArg_new(0, buf));
                    string_push_back(statment->mTitle, string_c_str(buf));
                    string_push_back(statment->mTitle, " ");
                }
            }
            else {
                vector_add(command->mArgs, sArg_new(0, buf));
                string_push_back(statment->mTitle, string_c_str(buf));
                string_push_back(statment->mTitle, " ");
            }
        }
        else {
            /// 空文字なら
            if(**p == '\'') {
                string_push_back(not_evaled, "''");
                (*p)+=2;
                vector_add(command->mArgs, sArg_new(0, buf));
                string_push_back(statment->mTitle, "'");
                string_push_back(statment->mTitle, " ");

                continue;
            }
            else if(**p == '"') {
                string_push_back(not_evaled, "\"\"");
                (*p)+=2;
                vector_add(command->mArgs, sArg_new(0, buf));
                string_push_back(statment->mTitle, "\"\"");
                string_push_back(statment->mTitle, " ");

                continue;
            }
        }
        
        /// クィックコマンド展開 ///
        if((**p == '@' && *(*p+1) == '(') || (**p == '@' && *(*p+1) == '@' && *(*p+2) == '(')) {
            BOOL double_;
            if(*(*p+1) == '@') {
                double_ = TRUE;
                string_push_back2(not_evaled, **p);
                (*p)++;
            }
            else {
                double_ = FALSE;
            }

            string_push_back2(not_evaled, **p); // @
            (*p)++;

            string_obj* str = STRING_NEW("");
            int read_end_of_statment = -1;
            if(!read_until_next_close_character('(', ')', p, str, &read_end_of_statment, FALSE, sname, sline)) {
                return FALSE;
            }
            string_push_back(not_evaled, string_c_str(str));
            string_push_back(statment->mTitle, string_c_str(str));

            string_erase(str, 0, 1);
            string_trunc(str, string_length(str)-1);

            sStatments* statments = STATMENTS_NEW();
            if(!parse(string_c_str(str), sname, sline, statments))
            {
                return FALSE;
            }

            if(double_) {
                vector_add(command->mArgs, sArg_new(4, statments));
            }
            else {
                vector_add(command->mArgs, sArg_new(2, statments));
            }
        }
        else if(**p == '>') {
            /// プロセス置換 ///
            if(*(*p+1) == '(') {
                string_push_back2(not_evaled, **p);
                (*p)++;

                string_obj* cmd = STRING_NEW("");
                if(!read_until_next_close_character('(', ')', p, cmd, read_end_of_statment, FALSE, sname, sline)) {
                    return FALSE;
                }
                string_push_back(not_evaled, string_c_str(cmd));

                string_erase(cmd, 0, 1); // (を取る
                string_trunc(cmd, string_length(cmd)-1); // )をとる

                sStatments* statments = STATMENTS_NEW();
                if(!parse(string_c_str(cmd), sname, sline, statments)) {
                    return FALSE;
                }

                vector_add(command->mArgs, sArg_new(8, statments));
            }
            /// リダイレクト
            else if(*(*p+1) == '>') {
                string_push_back(not_evaled, ">>");
                (*p)+=2;
                int fd = 1;
                enum eRedirect type = kRedirectAppend;

                skip_spaces(p, not_evaled);

                string_obj* fname = STRING_NEW("");
                if(!read_one_argument(p, fname, not_evaled, read_end_of_statment, NULL, sname, sline))
                {
                    return FALSE;
                }

                // クィックコマンド展開は置けない
                if(string_c_str(fname)[0] == 0 && (**p == '@' && *(*p+1) == '(' || **p == '@' && *(*p+1) == '@' && *(*p+2) == '(')) {
                    err_msg("can't put a @ command expression after redirect", sname, *sline);
                    return FALSE;
                }

                vector_add(command->mRedirects
                        , sRedirect_new(type, fd, string_c_str(fname)));

            }
            /// リダイレクト
            else {
                string_push_back(not_evaled, ">");
                (*p)++;
                int fd = 1;
                enum eRedirect type = kRedirectOverwrite;

                skip_spaces(p, not_evaled);

                string_obj* fname = STRING_NEW("");
                if(!read_one_argument(p, fname, not_evaled, read_end_of_statment, NULL, sname, sline))
                {
                    return FALSE;
                }

                // クィックコマンド展開は置けない
                if(string_c_str(fname)[0] == 0 && (**p == '@' && *(*p+1) == '(' || **p == '@' && *(*p+1) == '@' && *(*p+2) == '(')) {
                    err_msg("can't put a @ command expression after redirect", sname, *sline);
                    return FALSE;
                }

                vector_add(command->mRedirects
                        , sRedirect_new(type, fd, string_c_str(fname)));

            }
        }
        else if(**p == '<') {
            /// ヘアドキュメント
            if(*(*p+1) == '<') {
                string_push_back2(not_evaled, **p);
                (*p)++;
                string_push_back2(not_evaled, **p);
                (*p)++;

                skip_spaces(p, not_evaled);

                BOOL expand_env = TRUE;

                BOOL squote = FALSE;
                BOOL dquote = FALSE;
                if(**p == '\'') {
                    squote = TRUE;
                    string_push_back2(not_evaled, **p);
                    (*p)++;
                }
                else if(**p == '"') {
                    dquote = TRUE;
                    string_push_back2(not_evaled, **p);
                    (*p)++;
                }

                string_obj* name = STRING_NEW("");
                while(**p) {
                    if(**p == '\n') {
                        string_push_back2(not_evaled, **p);
                        (*p)++;
                        break;
                    }
                    else if(**p == ' ' || **p == '\t') {
                        skip_spaces(p, not_evaled);
                        if(**p != '\n') {
                            err_msg("invalid here document name", sname, *sline);
                            return FALSE;
                        }
                        string_push_back2(not_evaled, **p);
                        (*p)++;
                        break;
                    }
                    else if(**p >='A' && ** p <= 'Z' || **p >='a' && **p <='z')
                    {
                        string_push_back2(name, **p);
                        string_push_back2(not_evaled, **p);
                        (*p)++;
                    }
                    else if(dquote && **p == '"') {
                        expand_env = TRUE;
                        dquote = FALSE;
                        string_push_back2(not_evaled, **p);
                        (*p)++;
                        skip_spaces(p, not_evaled);
                        if(**p != '\n') {
                            err_msg("invalid here document name", sname, *sline);
                            return FALSE;
                        }
                        string_push_back2(not_evaled, **p);
                        (*p)++;
                        break;
                    }
                    else if(squote && **p == '\'') {
                        expand_env = FALSE;
                        squote = FALSE;
                        string_push_back2(not_evaled, **p);
                        (*p)++;
                        skip_spaces(p, not_evaled);
                        if(**p != '\n') {
                            err_msg("invalid here document name", sname, *sline);
                            return FALSE;
                        }
                        string_push_back2(not_evaled, **p);
                        (*p)++;
                        break;
                    }
                    else {
                        err_msg("invalid here document name", sname, *sline);
                        return FALSE;
                    }
                }

                if(dquote || squote) {
                    err_msg("invalid here document name", sname, *sline);
                    return FALSE;
                }
                
                if(string_c_str(name)[0] == 0) {
                    err_msg("invalid here document name", sname, *sline);
                    return FALSE;
                }

                string_obj* text = STRING_NEW("");
                string_obj* line = STRING_NEW("");
                while(**p) {
                    if(**p == '\n') {
                        string_push_back2(not_evaled, **p);
                        (*p)++;
                        string_push_back(text, string_c_str(line));
                        string_push_back2(text, '\n');
                        string_put(line, "");
                    }
                    /// コマンド展開 ///
                    else if(expand_env && **p == '$' && *(*p+1) == '(' || **p == '$' && *(*p+1) == '$' && *(*p+2) == '(') 
                    {
                        if(*(*p+1) == '$') {
                            string_push_back2(not_evaled, **p); // $
                            (*p)++;
                            string_push_back2(not_evaled, **p); // $
                            (*p)++;
                        }
                        else {
                            string_push_back2(not_evaled, **p); // $
                            (*p)++;
                        }

                        if(!read_until_next_close_character('(', ')', p, not_evaled, read_end_of_statment, TRUE, sname, sline)) {
                            return FALSE;
                        }

                        *read_end_of_statment = 1;
                    }
                    else {
                        string_push_back2(line, **p);
                        string_push_back2(not_evaled, **p);
                        (*p)++;

                        if(strstr(string_c_str(line), string_c_str(name)) == string_c_str(line))
                        {
                            break;
                        }
                    }
                }

                vector_add(command->mArgs, sArg_new(7, text));
            }
            /// プロセス置換 ///
            else if(*(*p+1) == '(') {
                string_push_back2(not_evaled, **p);
                (*p)++;

                string_obj* cmd = STRING_NEW("");
                if(!read_until_next_close_character('(', ')', p, cmd, read_end_of_statment, FALSE, sname, sline)) {
                    return FALSE;
                }
                string_push_back(not_evaled, string_c_str(cmd));

                string_erase(cmd, 0, 1); // (を取る
                string_trunc(cmd, string_length(cmd)-1); // )をとる

                sStatments* statments = STATMENTS_NEW();
                if(!parse(string_c_str(cmd), sname, sline, statments)) {
                    return FALSE;
                }

                vector_add(command->mArgs, sArg_new(6, statments));
            }
            /// リダイレクト
            else {
                (*p)++;
                string_push_back(not_evaled, "<");
                int fd = 0;
                enum eRedirect type = kRedirectInput;

                skip_spaces(p, not_evaled);

                string_obj* fname = STRING_NEW("");
                if(!read_one_argument(p, fname, not_evaled, read_end_of_statment, NULL, sname, sline))
                {
                    return FALSE;
                }

                // クィックコマンド展開は置けない
                if(string_c_str(fname)[0] == 0 && (**p == '@' && *(*p+1) == '(' || **p == '@' && *(*p+1) == '@' && *(*p+2) == '(')) {
                    err_msg("can't put a @ command expression after redirect", sname, *sline);
                    return FALSE;
                }

                vector_add(command->mRedirects
                        , sRedirect_new(type, fd, string_c_str(fname)));

            }
        }
        /// リダイレクト ///
        else if(**p == '%') {
            if(*(*p+1) == '>' && *(*p+2) == '>') {
                string_push_back(not_evaled, "%>>");
                break;
            }
            else if(*(*p+1) == '2' && *(*p+2) == '>' && *(*p+3) == '>') 
            {
                string_push_back(not_evaled, "%2>>");
                (*p)+=4;
                int fd = 2;
                enum eRedirect type = kRedirectAppend;

                skip_spaces(p, not_evaled);

                string_obj* fname = STRING_NEW("");
                if(!read_one_argument(p, fname, not_evaled, read_end_of_statment, NULL, sname, sline))
                {
                    return FALSE;
                }

                // クィックコマンド展開は置けない
                if(string_c_str(fname)[0] == 0 && (**p == '@' && *(*p+1) == '(' || **p == '@' && *(*p+1) == '@' && *(*p+2) == '(')) {
                    err_msg("can't put a @ command expression after redirect", sname, *sline);
                    return FALSE;
                }

                vector_add(command->mRedirects
                        , sRedirect_new(type, fd, string_c_str(fname)));

            }
            else if(*(*p+1) == '2' && *(*p+2) == '>') {
                string_push_back(not_evaled, "%2>");
                (*p)+=3;
                int fd = 2;
                enum eRedirect type = kRedirectOverwrite;

                skip_spaces(p, not_evaled);

                string_obj* fname = STRING_NEW("");
                if(!read_one_argument(p, fname, not_evaled, read_end_of_statment, NULL, sname, sline))
                {
                    return FALSE;
                }

                // クィックコマンド展開は置けない
                if(string_c_str(fname)[0] == 0 && (**p == '@' && *(*p+1) == '(' || **p == '@' && *(*p+1) == '@' && *(*p+2) == '(')) {
                    err_msg("can't put a @ command expression after redirect", sname, *sline);
                    return FALSE;
                }

                vector_add(command->mRedirects
                        , sRedirect_new(type, fd, string_c_str(fname)));

            }
            else if(*(*p+1) == '>') {
                string_push_back(not_evaled, "%>");
                (*p)+=2;

                int fd = 0;
                enum eRedirect type = kRedirectErrAndOutput;

                skip_spaces(p, not_evaled);

                string_obj* fname = STRING_NEW("");
                if(!read_one_argument(p, fname, not_evaled, read_end_of_statment, NULL, sname, sline))
                {
                    return FALSE;
                }

                // クィックコマンド展開は置けない
                if(string_c_str(fname)[0] == 0 && (**p == '@' && *(*p+1) == '(' || **p == '@' && *(*p+1) == '@' && *(*p+2) == '(')) {
                    err_msg("can't put a @ command expression after redirect", sname, *sline);
                    return FALSE;
                }

                vector_add(command->mRedirects
                        , sRedirect_new(type, fd, string_c_str(fname)));

            }
            else {
                char buf[128];
                snprintf(buf, 128, "unexpected token1 -->%c\n", **p);
                err_msg(buf, sname, *sline);
                return FALSE;
            }
        }
        /// 空白 ///
        else if(**p == ' ' || **p == '\t') {
            skip_spaces(p, not_evaled);
        }
        /// サブシェル ///
        else if(**p == '(') {
            if(vector_size(command->mArgs) == 0) {
                string_obj* str = STRING_NEW("");
                int read_end_of_statment = -1;
                if(!read_until_next_close_character('(', ')', p, str, &read_end_of_statment, FALSE, sname, sline)) {
                    return FALSE;
                }
                string_push_back(not_evaled, string_c_str(str));
                string_push_back(statment->mTitle, string_c_str(str));

                string_erase(str, 0, 1);
                string_trunc(str, string_length(str)-1);

                sStatments* statments = STATMENTS_NEW();
                if(!parse(string_c_str(str), sname, sline, statments))
                {
                    return FALSE;
                }

                command->mKind= kSubshell;
                vector_add(command->mArgs, sArg_new(0, STRING_NEW("subshell")));
                command->mExtra = sSubshell_new(statments);
            }
            else {
                char buf[128];
                snprintf(buf, 128, "unexpected token2 -->%c\n", **p);
                err_msg(buf, sname, *sline);
                return FALSE;
            }
        }
        /// ブロック ///
        else if(**p == '{') {
            string_obj* str = STRING_NEW("");
            vector_obj* args = VECTOR_NEW(10);
            if(!read_block(p, str, args, not_evaled, FALSE, sname, sline)) {
                return FALSE;
            }

            sStatments* block = STATMENTS_NEW();
            if(!parse(string_c_str(str), sname, sline, block))
            {
                return FALSE;
            }

            int count = 0;
            int i;
            for(i=0; i<vector_size(command->mArgs); i++) {
                sArg* arg = vector_item(command->mArgs, i);
                if(arg->mKind != 5) {
                    count++;
                }
            }

            vector_add(command->mArgs, sArg_new(5, sBlock_new(block, args, count-1)));

            string_push_back(statment->mTitle, string_c_str(str));
            string_push_back(statment->mTitle, " ");

            skip_spaces(p, not_evaled);
        }
        else {
            break;
        }
    }

    return TRUE;
}

/// 一文読み込み
static BOOL read_statment(char**p, sStatment* statment, string_obj* not_evaled, int* read_end_of_statment, BOOL block_termination, char* sname, int* sline)
{
    /// 反転 ///
    if(**p == '!') {
        string_push_back2(not_evaled, '!');
        (*p)++;
        statment->mRCodeReverse = TRUE;

        skip_spaces(p, not_evaled);
    }

    /// グローバルパイプ, STDINパイプ ///
    if(**p == '|' && *(*p+1) == '>') {
        string_push_back2(not_evaled, **p);
        (*p)++;
        string_push_back2(not_evaled, **p);
        (*p)++;

        statment->mGlobalPipeIn = 1;

        skip_spaces(p, not_evaled);
    }
    else if(**p == '|' && *(*p+1) >= '0' && *(*p+1) <= '9' && *(*p+2) == '>') {
        string_push_back2(not_evaled, **p);
        (*p)++;
        int num = **p - '0';
        string_push_back2(not_evaled, **p);
        (*p)++;
        string_push_back2(not_evaled, **p);
        (*p)++;

        statment->mGlobalPipeIn = 4;
        statment->mGlobalPipeInNum = num;

        skip_spaces(p, not_evaled);
    }
    else if(**p == '|' && *(*p+1) != '|') {
        string_push_back2(not_evaled, **p);
        (*p)++;

        skip_spaces(p, not_evaled);

        statment->mSTDINPipe = TRUE;
    }

    while(**p) {
        sCommand* command = COMMAND_NEW();
        if(!read_command(p, command, statment, not_evaled, read_end_of_statment, block_termination, sname, sline)) 
        {
            return FALSE;
        }

        /// 文に追加 ///
        if(*read_end_of_statment == -1 
            && vector_size((command)->mArgs) > 0) 
        {
            /// コマンド名からコマンド種別を得る ///
            sArg* arg = vector_item(command->mArgs, 0);
            if(arg->mKind == 0) {
                command->mKind = get_command_kind(string_c_str(arg->mBody));
            }
            else {
                err_msg("invalid command name", sname, *sline);
                return FALSE;
            }

            vector_add(statment->mCommands, command);
        }

        if(**p == '|' && *(*p+1) != '>' && (*(*p+1) < '0' || *(*p+1) > '9') && *(*p+1) != '|')
        {
            string_push_back(not_evaled, "|");
            (*p)++;
        }
        else {
            break;
        }
    }

    return TRUE;
}

/// エラー時はgErrMsgにエラーメッセージがあるので出力してください
BOOL parse(char* cmdline, char* sname, int* sline, sStatments* statments)
{
    char* p = cmdline;

    while(*p) {
        /// 行前のスペースを削除 ///
        while(*p == ' ' || *p == '\t' || *p == '\n' && (*sline)++) { p++; }

        /// 開始 ///
        sStatment* statment = STATMENT_NEW();
        string_obj* not_evaled = STRING_NEW("");
        int read_end_of_statment = -1;
            // このフラグに行番号が入っていたら
            // コマンドライン展開があったので行末まで読み込む
        int line_of_head = *sline;

        if(!read_statment(&p, statment, not_evaled, &read_end_of_statment, FALSE, sname, sline))
        {
            return FALSE;
        }

        /// グローバルパイプ
        if(*p == '|' && *(p+1) >= '0' && *(p+1) <= '9' && *(p+2) == '>') {
            string_push_back(not_evaled, "|");
            p++;
            int num = *p - '0';
            string_push_back2(not_evaled, *p);
            p++;
            string_push_back2(not_evaled, *p);
            p++;

            if(*p == '>') {
                string_push_back(not_evaled, ">");
                p++;
                statment->mGlobalPipeOut = 0x14;
                statment->mGlobalPipeOutNum = num;
            }
            else {
                statment->mGlobalPipeOut = 4;
                statment->mGlobalPipeOutNum = num;
            }

            skip_spaces(&p, not_evaled);
        }

        else if(*p == '|' && *(p+1) == '>') {
            string_push_back(not_evaled, "|");
            p++;
            string_push_back(not_evaled, ">");
            p++;

            if(*p == '>') {
                string_push_back(not_evaled, ">");
                p++;
                statment->mGlobalPipeOut = 0x11;
            }
            else {
                statment->mGlobalPipeOut = 0x01;
            }

            skip_spaces(&p, not_evaled);
        }

        /// 文の区切り
        if(*p == '&' && *(p+1) == '&') {
            string_push_back(not_evaled, "&&");
            p+=2;
            statment->mTerminated = kTAndAnd;
            statment->mLine = line_of_head;
            string_put(statment->mFName, sname);

            skip_spaces(&p, not_evaled);
        }
        else if(*p == '|' && *(p+1) == '|') {
            string_push_back(not_evaled, "||");
            p+=2;

            statment->mTerminated = kTOrOr;
            statment->mLine = line_of_head;
            string_put(statment->mFName, sname);

            skip_spaces(&p, not_evaled);
        }
        else if(*p == '&') {
            string_push_back2(not_evaled, '&');
            p++;

            statment->mBackground = TRUE;
            statment->mTerminated = kTNormal;
            statment->mLine = line_of_head;
            string_put(statment->mFName, sname);

            skip_spaces(&p, not_evaled);
        }
        else if(*p == '\n') {
            statment->mTerminated = kTNormal;
            statment->mLine = line_of_head;
            string_put(statment->mFName, sname);

            (*sline)++;
            string_push_back2(not_evaled, '\n');
            p++; 
            skip_spaces(&p, not_evaled);
        }
        else if(*p == ';') {
            statment->mTerminated = kTNormal;
            statment->mLine = line_of_head;
            string_put(statment->mFName, sname);

            string_push_back2(not_evaled, ';');
            p++;
            skip_spaces(&p, not_evaled);
        }
        else if(*p == 0) {
            statment->mTerminated = kTNormal;
            statment->mLine = line_of_head;
            string_put(statment->mFName, sname);
        }
        else {
            char buf[128];
            snprintf(buf, 128, "unexpected token3 -->%c\n", *p);
            err_msg(buf, sname, *sline);
            return FALSE;
        }

        /// コマンド展開があったなら
        if(read_end_of_statment != -1) {
            statment->mNotEvaled = not_evaled;
        }

        // コマンドがあるか、コマンド展開があったら複文に文を追加
        if(vector_size(statment->mCommands) > 0 || statment->mNotEvaled != NULL) {
            vector_add(statments->mStatments, statment);
        }
    }

    return TRUE;
}

// 文字列をクォートして返す
static BOOL expand_env_get_quoted_str(char* fname, string_obj* quoted_fname, char* sname, int* sline)
{
    char* p = fname;
    while(*p) {
        if(gKitutukiSigInt) {
            gKitutukiSigInt = FALSE;
            err_msg("signal interrupt", sname, *sline);
            return FALSE;
        }
        if(*p == '\n') {
            string_push_back(quoted_fname, "\\n");
            p++;
        }
        else if(*p == '\r') {
            string_push_back(quoted_fname, "\\r");
            p++;
        }
        else if(*p == '\t') {
            string_push_back(quoted_fname, "\\t");
            p++;
        }
        else if(*p == '\a') {
            string_push_back(quoted_fname, "\\a");
            p++;
        }
        else if(*p==' ' || *p=='*' || *p=='&' || *p=='~'
                || *p=='#' || *p =='$' || *p=='(' || *p==')'
                || *p=='\\' || *p=='|' || *p=='[' || *p==']' || *p=='{' 
                || *p=='}' || *p==';'
                || *p=='\'' || *p=='"' || *p=='<' || *p=='>' || *p=='?' 
                || *p=='%' || *p=='@')
        {
            string_push_back2(quoted_fname, '\\');
            string_push_back2(quoted_fname, *p++);
        }
        else {
            string_push_back2(quoted_fname, *p++);
        }
    }

    return TRUE;
}

BOOL expand_env(char* p, string_obj* out, sRFd* rfd, char* sname, int* sline, sObject* object)
{
    BOOL squote = FALSE;
    BOOL dquote = FALSE;

    while(*p) {
        if(gKitutukiSigInt) {
            err_msg("signal interrupt", sname, *sline);
            gKitutukiSigInt = FALSE;
            return FALSE;
        }

        /// クォート ///
        if(!squote && !dquote && *p == '\\') {
            string_push_back2(out, *p++);
            if(*p == 0) {
                err_msg("unexpected end after quote", sname, *sline);
                return FALSE;
            }

            string_push_back2(out, *p++);
        }
        /// シングルクォート ///
        else if(!dquote && *p == '\'') {
            squote = !squote;
            string_push_back2(out, *p++);
        }
        /// ダブルクォート ///
        else if(!squote && *p == '"') {
            dquote = !dquote;
            string_push_back2(out, *p++);
        }
        /// コマンド展開
        else if(!squote && *p == '$') {
            /// コマンド置換 $( ///
            if(*(p+1) == '$' && *(p+2) == '(' || *(p+1) == '(')
            {
                BOOL quote;
                if(*(p+1) == '$') {
                    p+=2;
                    quote = FALSE;
                }
                else {
                    p++;
                    quote = TRUE;
                }

                string_obj* str = STRING_NEW("");
                int read_end_of_statment = -1;
                if(!read_until_next_close_character('(', ')', &p, str, &read_end_of_statment, FALSE, sname, sline)) 
                {
                    return FALSE;
                }

                string_erase(str, 0, 1); // (をとる
                string_trunc(str, string_length(str)-1); // )をとる

                string_obj* ret = STRING_NEW("");
                char* fname = MALLOC(strlen(sname) + 256);
                snprintf(fname, strlen(sname) + 256, "%s %d: expand command", sname, *sline);
                int rcode = saphire_shell5(ret, string_c_str(str), fname, rfd, object);
                if(rcode < 0) {
                    return FALSE;
                }

                if(quote) {
                    string_obj* quoted = STRING_NEW("");
                    if(!expand_env_get_quoted_str(string_c_str(ret), quoted, sname, sline))
                    {
                        return FALSE;
                    }
                    string_push_back(out, string_c_str(quoted));
                }
                else {
                    string_push_back(out, string_c_str(ret));
                }
            }
            else {
                string_push_back2(out, *p++);
            }
        }
        else if(squote || dquote) {
            if(*p == 0) {
                err_msg("require ' or \"", sname, *sline);
                return FALSE;
            }

            string_push_back2(out, *p++);
        }

        /// ブロックは飛ばす
        else if(*p == '{') {
            string_obj* dummy = STRING_NEW("");
            if(!read_block(&p, dummy, NULL, out, FALSE, sname, sline)) {
                return FALSE;
            }
        }
        /// サブシェル、条件式は飛ばす
        else if(*p == '(') {
            int read_end_of_statment = -1;
            if(!read_until_next_close_character('(', ')', &p, out, &read_end_of_statment, FALSE, sname, sline))
            {
                return FALSE;
            }
        }
        else {
            string_push_back2(out, *p++);
        }
    }

    return TRUE;
}

// コメントも除去する
// 読み込みだけ。この後パースする
BOOL delete_comments(char* p, string_obj* cmdline2, char* separator, char* sname, int* sline)
{
    BOOL squote = FALSE;
    BOOL dquote = FALSE;

    while(*p) {
        if(!dquote && !squote && *p == '\\') {
            string_push_back2(cmdline2, *p++);
            string_push_back2(cmdline2, *p++);
        }
        else if(!dquote && *p == '\'') {
            string_push_back2(cmdline2, *p++);
            squote = !squote;
        }
        else if(!squote && *p == '"') {
            string_push_back2(cmdline2, *p++);
            dquote = !dquote;
        }
        else if(squote || dquote) {
            if(*p == 0) {
                err_msg("require \" or \'", sname, *sline);
                return FALSE;
            }
            else {
                string_push_back2(cmdline2, *p++);
            }
        }
        /// コメント ///
        else if(*p == '#') {
            p++;

            while(*p) {
                if(*p == '\n') {
                    string_push_back2(cmdline2, '\n');
                    p++;

                    break;
                }
                else {
                    p++; 
                }
            }
        }
        else {
            string_push_back2(cmdline2, *p++);
        }
    }

    return TRUE;
}

// スクリプトファイルを読み込んでコンパイルして
// ファイルとして書き込む関数。
// エラー時はgErrMsgにエラーメッセージが入っている
BOOL saphire_compile(char* fname, char* out_fname)
{
    /// 読み込み ///
    string_obj* str = STRING_NEW("");

    int f = open(fname, O_RDONLY);
    if(f < 0) {
        return FALSE;
    }

    char buf[BUFSIZ];
    while(1) {
        int size = read(f, buf, BUFSIZ-1);
        if(size == 0) {
            break;
        }
        if(size < 0) {
            close(f);
            return FALSE;
        }

        buf[size] = 0;

        string_push_back(str, buf);
    }
    close(f);
    
    /// パース ///
    sStatments* statments = STATMENTS_NEW();

    int sline = 1;

    string_obj* str2 = STRING_NEW("");
    if(!delete_comments(string_c_str(str), str2, " ", fname, &sline)) {

        return FALSE;
    }

    sline = 1;

    if(!parse(string_c_str(str2), fname, &sline, statments)) {
        return FALSE;
    }
    

    /// コンパイル結果を保存 ///
    int fd = open(out_fname, O_WRONLY|O_TRUNC|O_CREAT, 0644);
    if(write(fd, gMagicNumber, strlen(gMagicNumber)) < 0) {
        perror("write");
        exit(1);
    }

    sStatments_save(statments, fd);

    (void)close(fd);

    return TRUE;
}

// オブジェクトファイル(コンパイルされたファイル)を読み込む関数。
// 実行もする。
// 戻り値が-1ならエラーメッセージがgErrMsgに入っているので出力してください
int saphire_load_obj(char* fname, sWFd* pipeout, sRFd* pipein, int pipeerr, vector_obj* parent_blocks)
{
    gKitutukiSigInt = FALSE;
    int fd = open(fname, O_RDONLY);
    if(fd < 0) {
        return -1;
    }
    
    char magic_number[64];
    if(read(fd, magic_number, strlen(gMagicNumber)) < 0) {
        perror("read");
        exit(1);
    }
    magic_number[strlen(gMagicNumber)] = 0;
    
    if(strcmp(magic_number, gMagicNumber) != 0) {
        return -1;
    }
    
    /// オブジェクトファイルをロード ///
    sStatments* statments = sStatments_load(fd);

    /// 実行 ///
    int return_ = FALSE;
    int break_ = FALSE;
    vector_obj* envs = VECTOR_NEW(10);
    int rcode = run(statments, fname
                , pipeout, pipein, pipeerr
                , &return_, TRUE, &break_, FALSE, parent_blocks, FALSE, envs, NULL, NULL, NULL);

    /// 解放 ///
    close(fd);

    return rcode;
}
